import math


# https://chatgpt.com/c/67b173a0-26cc-8005-9bb4-46bfbc1d53ee
def meters_to_degrees(
    lat: float,
    delta_in_m: float,
    direction: str,
) -> float:
    earth_radius = 6378137.0

    # Conversion factor from degrees to radians
    degrees_to_radians = math.pi / 180.0

    # Calculate the number of degrees latitude per meter
    meters_per_degree_latitude = earth_radius * degrees_to_radians

    # Calculate the number of degrees longitude per meter
    # The radius of a circle of latitude decreases with the cosine of the latitude
    meters_per_degree_longitude = meters_per_degree_latitude * math.cos(
        lat * degrees_to_radians
    )

    if direction == "lat":
        return delta_in_m / meters_per_degree_latitude

    if direction == "lon":
        return delta_in_m / meters_per_degree_longitude

    raise ValueError(f"'lat' or 'lon' expected, received '{direction}'!")



from roboflow import Roboflow

from blueness import module
from blue_options.elapsed_timer import ElapsedTimer

from roofai import NAME
from roofai.env import ROBOFLOW_API_KEY
from roofai.logger import logger


NAME = module.name(__file__, NAME)

list_of_project_types = [
    "object-detection",
    "single-label-classification",
    "multi-label-classification",
    "instance-segmentation",
    "semantic-segmentation",
]


# https://docs.roboflow.com/api-reference/images/upload-api#upload-to-a-new-project
def create_project(
    project_name: str,
    project_description: str,
    project_type: str = "semantic-segmentation",
    project_license: str = "MIT",
    recreate: bool = False,
) -> bool:
    if project_type not in list_of_project_types:
        logger.error(f"{project_type}: project type not found.")
        return False

    rf = Roboflow(api_key=ROBOFLOW_API_KEY)

    if not recreate and any(
        existing_project_name.startswith(f"kamangir/{project_name}")
        for existing_project_name in rf.workspace().projects()
    ):
        logger.info(f"✅  {project_name}")
        return True

    logger.info(
        "{}.create_project({}/{}): {}".format(
            NAME,
            project_type,
            project_name,
            project_description,
        )
    )

    timer = ElapsedTimer()
    try:
        _ = rf.workspace().create_project(
            project_name=project_name,
            project_license=project_license,
            project_type=project_type,
            annotation=project_description,
        )
    except Exception as e:
        logger.error(e)
        return False
    timer.stop()
    logger.info(f"took {timer.elapsed_pretty()}")

    return True



from abcli.tests.test_env import test_abcli_env
from blue_objects.tests.test_env import test_blue_objects_env

from roofai import env


def test_required_env():
    test_abcli_env()
    test_blue_objects_env()


def test_blue_plugin_env():
    assert env.GOOGLE_MAPS_API_KEY

    assert env.TEST_roofAI_ingest_AIRS_v1
    assert env.TEST_roofAI_ingest_AIRS_v2
    assert env.TEST_roofAI_ingest_CamVid_v1

    assert env.TEST_roofAI_semseg_model_AIRS_full_v1
    assert env.TEST_roofAI_semseg_model_AIRS_full_v2
    assert env.TEST_roofAI_semseg_model_CamVid_v1

    assert env.ROOFAI_AIRS_CACHE_OBJECT_NAME

    assert isinstance(env.ROOFAI_TEST_GOOGLE_MAPS_LAT, float)
    assert isinstance(env.ROOFAI_TEST_GOOGLE_MAPS_LON, float)

    assert isinstance(env.ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LAT, float)
    assert isinstance(env.ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LON, float)

    assert env.ROBOFLOW_API_KEY

    assert env.ROOFAI_DEFAULT_GOOGLE_MAPS_MODEL



import os
import matplotlib.pyplot as plt
import time
from typing import List
import torch
from torch.utils.data import DataLoader
import segmentation_models_pytorch as smp
from segmentation_models_pytorch import utils

from blue_options import string
from blue_objects import file, path
from blue_objects.graphics.signature import sign_filename

from roofai.semseg.augmentation import (
    get_training_augmentation,
    get_validation_augmentation,
    get_preprocessing,
)
from roofai.dataset.classes import RoofAIDataset
from roofai.semseg.dataloader import Dataset
from roofai.semseg.model import SemSegModel
from roofai.semseg.utils import visualize
from roofai.semseg import Profile
from roofai.host import signature
from roofai.logger import logger


class SemSegModelTrainer:
    def __init__(
        self,
        dataset_path: str,
        model_path: str,
        classes: List[str],
        in_notebook: bool = False,
        profile: Profile = Profile.VALIDATION,
    ):
        self.dataset = RoofAIDataset(dataset_path)

        self.dataset_path = self.dataset.dataset_path

        self.model_path = model_path
        path.create(self.model_path, log=True)

        self.in_notebook = in_notebook
        self.profile = profile

        logger.info(
            "{}: {} -{}-> {}".format(
                self.__class__.__name__,
                path.name(self.dataset_path),
                self.profile,
                path.name(self.model_path),
            )
        )

        self.x_train_dir = os.path.join(self.dataset_path, "train")
        self.y_train_dir = os.path.join(self.dataset_path, "trainannot")

        self.x_valid_dir = os.path.join(self.dataset_path, "val")
        self.y_valid_dir = os.path.join(self.dataset_path, "valannot")

        self.x_test_dir = os.path.join(self.dataset_path, "test")
        self.y_test_dir = os.path.join(self.dataset_path, "testannot")

        dataset = Dataset(
            self.x_train_dir,
            self.y_train_dir,
            classes=classes,
            count=self.profile.data_count,
        )

        image, mask = dataset[0]  # get some sample
        visualize(
            {
                "image": image,
                "mask": mask.squeeze(),
            },
            in_notebook=in_notebook,
            filename=os.path.join(model_path, "dataset.png"),
        )

        #### Visualize resulted augmented images and masks
        augmented_dataset = Dataset(
            self.x_train_dir,
            self.y_train_dir,
            augmentation=get_training_augmentation(),
            classes=classes,
            count=self.profile.data_count,
        )

        # same image with different random transforms
        for i in range(1 if profile == Profile.VALIDATION else 3):
            image, mask = augmented_dataset[0]
            visualize(
                {
                    "image": image,
                    "mask": mask.squeeze(-1),
                },
                in_notebook=in_notebook,
                filename=os.path.join(model_path, f"augmented_dataset-{i:05d}.png"),
            )

    def train(
        self,
        classes,
        encoder_name="se_resnext50_32x4d",
        encoder_weights="imagenet",
        activation="sigmoid",  # could be None for logits or 'softmax2d' for multi-class segmentation
        device="cpu",  # 'cuda'
        in_notebook: bool = False,
        epoch_count: int = -1,
    ):
        if epoch_count == -1:
            epoch_count = self.profile.epoch_count

        logger.info(
            "{}.train{} -{}:{}-> {}[{}] {} epoch(s)".format(
                self.__class__.__name__,
                device,
                activation,
                encoder_name,
                encoder_weights,
                ",".join(classes),
                epoch_count,
            )
        )

        start_time = time.time()

        # create segmentation model with pretrained encoder
        model = smp.FPN(
            encoder_name=encoder_name,
            encoder_weights=encoder_weights,
            classes=len(classes),
            activation=activation,
        ).to(device)

        preprocessing_fn = smp.encoders.get_preprocessing_fn(
            encoder_name,
            encoder_weights,
        )

        train_dataset = Dataset(
            self.x_train_dir,
            self.y_train_dir,
            augmentation=get_training_augmentation(),
            preprocessing=get_preprocessing(preprocessing_fn),
            classes=classes,
            count=self.profile.data_count,
        )

        valid_dataset = Dataset(
            self.x_valid_dir,
            self.y_valid_dir,
            augmentation=get_validation_augmentation(),
            preprocessing=get_preprocessing(preprocessing_fn),
            classes=classes,
            count=self.profile.data_count,
        )

        train_loader = DataLoader(
            train_dataset,
            batch_size=8,
            shuffle=True,
            num_workers=0,
        )
        valid_loader = DataLoader(
            valid_dataset,
            batch_size=1,
            shuffle=False,
            num_workers=0,
        )

        # Dice/F1 score - https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
        # IoU/Jaccard score - https://en.wikipedia.org/wiki/Jaccard_index

        loss = utils.losses.DiceLoss()
        metrics = [
            utils.metrics.IoU(threshold=0.5),
        ]

        optimizer = torch.optim.Adam(
            [
                dict(params=model.parameters(), lr=0.0001),
            ]
        )

        # create epoch runners
        # it is a simple loop of iterating over dataloader`s samples
        train_epoch = smp.utils.train.TrainEpoch(
            model,
            loss=loss,
            metrics=metrics,
            optimizer=optimizer,
            device=device,
            verbose=True,
        )

        valid_epoch = smp.utils.train.ValidEpoch(
            model,
            loss=loss,
            metrics=metrics,
            device=device,
            verbose=True,
        )

        max_score = 0
        model_filename = os.path.join(self.model_path, "model.pth")
        epic_logs = {}
        epic_list = range(0, epoch_count)
        for i in epic_list:
            logger.info("epoch: #{}/{}".format(i + 1, epoch_count))
            train_logs = train_epoch.run(train_loader)
            valid_logs = valid_epoch.run(valid_loader)

            epic_logs[i] = {
                "train": {key_: float(value_) for key_, value_ in train_logs.items()},
                "valid": {key_: float(value_) for key_, value_ in valid_logs.items()},
            }

            # do something (save model, change lr, etc.)
            if max_score < valid_logs["iou_score"]:
                max_score = valid_logs["iou_score"]
                torch.save(model, model_filename)
                logger.info("-> {}".format(model_filename))

            if i == 25:
                optimizer.param_groups[0]["lr"] = 1e-5
                print("Decrease decoder learning rate to 1e-5!")

        elapsed_time = time.time() - start_time
        logger.info(
            "took {}".format(
                string.pretty_duration(
                    elapsed_time,
                    largest=True,
                )
            )
        )
        file.save_json(
            os.path.join(self.model_path, "model.json"),
            {
                "activation": activation,
                "classes": classes,
                "elapsed_time": elapsed_time,
                "encoder_name": encoder_name,
                "encoder_weights": encoder_weights,
                "epics": epic_logs,
            },
        )

        plt.figure(figsize=(10, 5))
        plt.subplot(1, 2, 1)
        for subset in "train,valid".split(","):
            plt.plot(
                epic_list,
                [epic_logs[epic][subset]["dice_loss"] for epic in epic_list],
                label=f"{subset} loss",
            )
        plt.xlabel(
            "epic | total: {}".format(
                string.pretty_duration(
                    elapsed_time,
                    short=True,
                    largest=True,
                )
            )
        )
        plt.ylabel("dice-loss")
        plt.grid(True)
        plt.title(path.name(self.model_path))
        plt.legend()
        plt.subplot(1, 2, 2)
        for subset in "train,valid".split(","):
            plt.plot(
                epic_list,
                [epic_logs[epic][subset]["iou_score"] for epic in epic_list],
                label=f"{subset} iou",
            )
        plt.xlabel("epic")
        plt.ylabel("iou score")
        plt.ylim(0, 1)
        plt.grid(True)
        plt.legend()
        filename = os.path.join(self.model_path, "train-summary.png")
        file.prepare_for_saving(filename)
        plt.savefig(filename)
        if in_notebook:
            plt.show()
        plt.close()

        semseg_model = SemSegModel(
            model_filename,
            device=device,
        )

        sign_filename(
            filename,
            header=[
                "dataset: {}".format(
                    [
                        thing
                        for thing in self.dataset_path.split(os.sep)
                        if thing not in ["SegNet-Tutorial", "CamVid"] and thing
                    ][-1]
                ),
                "model: {}".format(path.name(self.model_path)),
                semseg_model.signature,
                "took {}".format(
                    string.pretty_duration(
                        elapsed_time,
                        largest=True,
                        short=True,
                    )
                ),
            ],
            footer=signature(),
        )

        test_dataset = Dataset(
            self.x_test_dir,
            self.y_test_dir,
            augmentation=get_validation_augmentation(),
            preprocessing=get_preprocessing(preprocessing_fn),
            classes=classes,
            count=self.profile.data_count,
        )

        test_dataloader = DataLoader(test_dataset)

        # evaluate model on test set
        test_epoch = smp.utils.train.ValidEpoch(
            model=semseg_model.model,
            loss=loss,
            metrics=metrics,
            device=device,
        )

        test_epoch.run(test_dataloader)  # logs

        # TODO: semseg_model.predict(...)

        return semseg_model



import pytest
from roofai.google_maps.semseg.utils import meters_to_degrees


@pytest.mark.parametrize(
    ["direction"],
    [
        ["lat"],
        ["lon"],
    ],
)
def test_google_maps_utils_meters_to_degrees(direction: str):
    lat = 49.2827
    delta_in_m = 100

    output = meters_to_degrees(
        lat=lat,
        delta_in_m=delta_in_m,
        direction=direction,
    )
    assert isinstance(output, float)


def test_google_maps_utils_meters_to_degrees_exception():
    lat = 49.2827
    delta_in_m = 100
    direction = "void"

    with pytest.raises(ValueError):
        meters_to_degrees(
            lat=lat,
            delta_in_m=delta_in_m,
            direction=direction,
        )



from typing import List

from blue_options.terminal import show_usage

from roofai.help.semseg import predict_options


def help_predict(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "lat=<lat>,lon=<lon>"

    args = [
        "[--address <address>]",
    ]

    return show_usage(
        [
            "@gmaps",
            "predict",
            f"[{options}]",
            "[{}]".format(predict_options(mono=mono)),
            "[<model-object-name>]",
            "[-|<prediction-object-name>]",
        ]
        + args,
        "<lat>, <lon> -predict-> <prediction-object-name>.",
        mono=mono,
    )



from roofai.help.google_maps.geocode import help_geocode
from roofai.help.google_maps.predict import help_predict
from roofai.help.google_maps.static_api import help_get_static_image

help_functions = {
    "geocode": help_geocode,
    "get_static_image": help_get_static_image,
    "predict": help_predict,
}



"""
Writing helper class for data extraction, transformation and preprocessing  
https://pytorch.org/docs/stable/data

Copied with minor modification from ../../notebooks/semseg.ipynb
"""

import os
import cv2
import numpy as np
from torch.utils.data import DataLoader
from torch.utils.data import Dataset as BaseDataset

from blue_objects import path

from roofai.dataset.classes import RoofAIDataset


class Dataset(BaseDataset):
    """CamVid Dataset. Read images, apply augmentation and preprocessing transformations.

    Args:
        images_dir (str): path to images folder
        masks_dir (str): path to segmentation masks folder
        class_values (list): values of classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing
            (e.g. normalization, shape manipulation, etc.)

    """

    def __init__(
        self,
        images_dir,
        masks_dir,
        classes=None,
        augmentation=None,
        preprocessing=None,
        count=-1,
    ):
        self.ids = os.listdir(images_dir)
        if count != -1:
            self.ids = self.ids[:count]

        self.images_fps = [os.path.join(images_dir, image_id) for image_id in self.ids]
        self.masks_fps = [os.path.join(masks_dir, image_id) for image_id in self.ids]

        # convert str names to class values on masks
        dataset = RoofAIDataset(path.parent(images_dir, 3))

        missing_classes = [cls for cls in classes if cls.lower() not in dataset.classes]
        assert not bool(
            missing_classes
        ), "{} class(es) are not found in the dataset: {}".format(
            len(missing_classes),
            ", ".join(missing_classes),
        )

        self.class_values = [dataset.classes.index(cls.lower()) for cls in classes]
        print(
            "{}: {} class(es): {}".format(
                self.__class__.__name__,
                len(self.class_values),
                ", ".join([str(value) for value in self.class_values]),
            )
        )
        assert self.class_values

        self.augmentation = augmentation
        self.preprocessing = preprocessing

        print(f"{self.__class__.__name__}: {len(self.ids)} item(s).")

    def __getitem__(self, i):
        # read data
        image = cv2.imread(self.images_fps[i])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        mask = cv2.imread(self.masks_fps[i], 0)

        # extract certain classes from mask (e.g. cars)
        masks = [(mask == v) for v in self.class_values]
        mask = np.stack(masks, axis=-1).astype("float")

        # apply augmentations
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        # apply preprocessing
        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        return image, mask

    def __len__(self):
        return len(self.ids)



from typing import List

from blue_options.terminal import show_usage, xtra


def help_browse(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "dev,sandbox"

    return show_usage(
        [
            "@google_earth",
            "browse",
            f"[{options}]",
        ],
        "browse google earth tools.",
        mono=mono,
    )


def help_fetch(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("dryrun,install,", mono=mono),
            "upload",
        ]
    )

    args = [
        "[--latitude=<49.279802>]",
        "[--longitude=<-123.115660>]",
        "[--screenSpaceError=<2>]",
        "[--width=<230>]",
        "[--height=<175>]",
        "[--zoom=<19>]",
    ]

    return show_usage(
        [
            "@google_earth",
            "fetch",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + args,
        "fetch from google earth.",
        mono=mono,
    )


help_functions = {
    "browse": help_browse,
    "fetch": help_fetch,
}



from enum import Enum, auto

chip_width = 480
chip_height = 384


class Profile(Enum):
    FULL = auto()
    DECENT = auto()
    QUICK = auto()
    DEBUG = auto()
    VALIDATION = auto()

    @property
    def data_count(self):
        return {
            Profile.FULL: -1,
            Profile.DECENT: 40,
            Profile.QUICK: 20,
            Profile.DEBUG: 2,
            Profile.VALIDATION: 1,
        }[self]

    @property
    def epoch_count(self):
        return {
            Profile.FULL: 40,
            Profile.DECENT: 10,
            Profile.QUICK: 5,
            Profile.DEBUG: 3,
            Profile.VALIDATION: 1,
        }[self]



from blue_options import string

from roofai.roboflow.create import create_project


def test_roboflow_create_project():
    assert create_project(
        project_name="roofai-pytest-roboflow-create-project",
        project_description="created-by-pytest",
    )



import pytest
import os
import numpy as np

from blue_objects import objects

from roofai.env import TEST_roofAI_ingest_CamVid_v1
from roofai.dataset.classes import RoofAIDataset, MatrixKind, DatasetKind


@pytest.mark.parametrize(
    "dataset_kind, matrix_kind, subset_path, filename_and_extension",
    [
        (
            DatasetKind.AIRS,
            MatrixKind.IMAGE,
            "subset/image",
            "subset/image/record_id.tif",
        ),
        (
            DatasetKind.AIRS,
            MatrixKind.MASK,
            "subset/label",
            "subset/label/record_id.tif",
        ),
        (
            DatasetKind.CAMVID,
            MatrixKind.IMAGE,
            "subset",
            "subset/record_id.png",
        ),
        (
            DatasetKind.CAMVID,
            MatrixKind.MASK,
            "subsetannot",
            "subsetannot/record_id.png",
        ),
    ],
)
def test_MatrixKind(
    dataset_kind: DatasetKind,
    matrix_kind: MatrixKind,
    subset_path,
    filename_and_extension,
):
    assert matrix_kind.subset_path(dataset_kind, "subset") == subset_path

    assert (
        matrix_kind.filename_and_extension(dataset_kind, "subset", "record_id")
        == filename_and_extension
    )


@pytest.mark.parametrize(
    [
        "dataset_object_name",
        "one_liner_prefix",
        "subset",
        "index",
        "matrix_kind",
        "expected_suffix",
        "expected_shape",
    ],
    [
        (
            TEST_roofAI_ingest_CamVid_v1,
            "RoofAIDataset[kind:DatasetKind.CAMVID,source:CamVid](",
            "test",
            10,
            MatrixKind.IMAGE,
            "test/0001TP_008850.png",
            (360, 480, 3),
        ),
        (
            TEST_roofAI_ingest_CamVid_v1,
            "RoofAIDataset[kind:DatasetKind.CAMVID,source:CamVid](",
            "test",
            10,
            MatrixKind.MASK,
            "testannot/0001TP_008850.png",
            (360, 480),
        ),
        (
            "palisades-dataset-v1",
            "RoofAIDataset[kind:DatasetKind.DISTRIBUTED,source:catalog_query](",
            "train",
            0,
            MatrixKind.IMAGE,
            "datacube-maxar_open_data-WildFires-LosAngeles-Jan-2025-11-031311102213-103001010B9A1B00/11-031311102213-103001010B9A1B00-103001010B9A1B00-visual.tif",
            (17408, 17408, 3),
        ),
        (
            "palisades-dataset-v1",
            "RoofAIDataset[kind:DatasetKind.DISTRIBUTED,source:catalog_query](",
            "train",
            0,
            MatrixKind.MASK,
            "datacube-maxar_open_data-WildFires-LosAngeles-Jan-2025-11-031311102213-103001010B9A1B00/11-031311102213-103001010B9A1B00-103001010B9A1B00-visual-label.tif",
            (17408, 17408),
        ),
    ],
)
def test_RoofAIDataset(
    dataset_object_name,
    one_liner_prefix,
    subset,
    index,
    matrix_kind,
    expected_suffix,
    expected_shape,
):
    assert objects.download(dataset_object_name)

    dataset = RoofAIDataset(objects.object_path(dataset_object_name))

    assert dataset.one_liner.startswith(one_liner_prefix)

    assert dataset.dataset_path

    record_id = dataset.subsets[subset][index]
    assert record_id

    filename = dataset.get_filename(subset, record_id, matrix_kind, log=True)
    assert filename.endswith(expected_suffix)

    matrix = dataset.get_matrix(subset, record_id, matrix_kind, log=True)
    assert matrix.shape == expected_shape
    assert matrix.dtype == np.uint8



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit
from blue_objects import objects

from roofai import NAME
from roofai.semseg import Profile
from roofai.google_maps.api.geocoding import geocode
from roofai.google_maps.api.static import get as get_static_image
from roofai.google_maps.semseg.ingest import ingest_dataset
from roofai.google_maps.semseg.predict import predict
from roofai.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="get_static_image | geocode | ingest_dataset | predict",
)
parser.add_argument(
    "--lat",
    type=float,
    default=0,
)
parser.add_argument(
    "--lon",
    type=float,
    default=0,
)
parser.add_argument(
    "--filename",
    type=str,
    default="",
)
parser.add_argument(
    "--object_name",
    type=str,
    default="",
)
parser.add_argument(
    "--zoom",
    type=int,
    default=20,
)
parser.add_argument(
    "--maptype",
    type=str,
    default="satellite",
)
parser.add_argument(
    "--size",
    type=str,
    default="640x640",
)
parser.add_argument(
    "--address",
    type=str,
    default="",
)
parser.add_argument(
    "--verbose",
    type=int,
    default=0,
    help="0 | 1",
)
parser.add_argument(
    "--count",
    type=int,
    default=10,
)
parser.add_argument(
    "--model_object_name",
    type=str,
)
parser.add_argument(
    "--prediction_object_name",
    type=str,
)
parser.add_argument(
    "--device",
    type=str,
    default="cpu",
    help="cpu|cuda",
)
parser.add_argument(
    "--profile",
    type=str,
    default="VALIDATION",
    help="FULL|QUICK|VALIDATION",
)
args = parser.parse_args()

success = False
if args.task == "get_static_image":
    success, _, _ = get_static_image(
        lat=args.lat,
        lon=args.lon,
        filename=objects.path_of(
            object_name=args.object_name,
            filename=(
                args.filename
                if args.filename
                else "{}.png".format(
                    "{}-{:.06f}-{:.06f}-{:02d}-{}".format(
                        args.maptype,
                        args.lat,
                        args.lon,
                        args.zoom,
                        args.size,
                    ).replace(".", "-")
                )
            ),
        ),
        zoom=args.zoom,
        maptype=args.maptype,
        size=args.size,
    )
elif args.task == "geocode":
    success, _, _, _ = geocode(
        address=args.address,
        object_name=args.object_name,
        verbose=args.verbose == 1,
    )
elif args.task == "ingest_dataset":
    success = ingest_dataset(
        object_name=args.object_name,
        lat=args.lat,
        lon=args.lon,
        zoom=args.zoom,
        maptype=args.maptype,
        size=args.size,
        count=args.count,
    )
elif args.task == "predict":
    success, _, _, _ = predict(
        lat=args.lat,
        lon=args.lon,
        address=args.address,
        model_object_name=args.model_object_name,
        prediction_object_name=args.prediction_object_name,
        device=args.device,
        profile=Profile[args.profile],
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Union, Dict
import requests
from PIL import Image
import numpy as np
import io
import math

from blueness import module
from blue_options import string

from roofai import NAME
from roofai.env import GOOGLE_MAPS_API_KEY
from roofai.google_maps.semseg.utils import meters_to_degrees
from roofai.logger import logger

NAME = module.name(__file__, NAME)

base_url = "https://maps.googleapis.com/maps/api/staticmap?"


# https://developers.google.com/maps/documentation/maps-static/start
def get(
    lat: float,
    lon: float,
    filename: str = "",
    zoom: int = 20,
    maptype: str = "satellite",
    size: str = "640x640",
) -> Union[bool, np.ndarray, Dict]:
    image: np.ndarray = np.array(())

    params = {
        "center": f"{lat},{lon}",
        "zoom": zoom,
        "size": size,
        "maptype": maptype,
        "key": GOOGLE_MAPS_API_KEY,
    }
    logger.info(
        "{}.get: {}".format(
            NAME,
            ", ".join(
                [f"{key}:{value}" for key, value in params.items() if key != "key"]
            ),
        )
    )

    response = requests.get(base_url, params=params)

    if response.status_code != 200:
        logger.error(
            "status_code={}, text={}".format(
                response.status_code,
                response.text,
            )
        )
        return False, image, {}

    try:
        image = np.array(Image.open(io.BytesIO(response.content)).convert("RGB"))
    except Exception as e:
        logger.error(e)
        return False, image, {}
    logger.info("image: {}".format(string.pretty_shape_of_matrix(image)))

    if filename:
        try:
            with open(filename, "wb") as file:
                file.write(response.content)
        except Exception as e:
            logger.error(e)
            return False, image, {}

        logger.info(f"-> {filename}")

    # https://groups.google.com/g/google-maps-js-api-v3/c/hDRO4oHVSeM?pli=1
    gsd = 156543.03392 * math.cos(lat * np.pi / 180) / (2**zoom)
    logger.info(f"gsd: {gsd:.2f} m")
    metadata = {"gsd": gsd}

    chip_height, chip_width = [int(piece) for piece in size.split("x")]
    if image.shape[0] != chip_height or image.shape[1] != chip_width:
        logger.warning(
            "image size mismatch: {} != {}x{}".format(
                string.pretty_shape_of_matrix(image),
                chip_height,
                chip_width,
            )
        )

    chip_height_in_meters = gsd * chip_height
    chip_width_in_meters = gsd * chip_width

    chip_height_in_degrees = meters_to_degrees(
        lat=lat,
        delta_in_m=chip_width_in_meters,
        direction="lat",
    )
    chip_width_in_degrees = meters_to_degrees(
        lat=lat,
        delta_in_m=chip_width_in_meters,
        direction="lon",
    )

    metadata["size"] = {
        "px": [chip_height, chip_width],
        "m": [chip_height_in_meters, chip_width_in_meters],
        "deg": [chip_height_in_degrees, chip_width_in_degrees],
    }
    logger.info(
        " = ".join(
            [
                "image size",
                "{} px x {} px".format(
                    chip_height,
                    chip_width,
                ),
                "{:,.2f} m x {:,.2f} m".format(
                    chip_height_in_meters,
                    chip_width_in_meters,
                ),
                "{:.5f} deg x {:.5f} deg ".format(
                    chip_height_in_degrees,
                    chip_width_in_degrees,
                ),
            ]
        )
    )

    return True, image, metadata



import os
from roboflow import Roboflow
from tqdm import tqdm
import glob
import matplotlib.pyplot as plt
import numpy as np

from blueness import module
from blue_options.elapsed_timer import ElapsedTimer
from blue_objects import objects, path, file
from blue_objects.metadata import post_to_object

from roofai import NAME, fullname
from roofai.env import ROBOFLOW_API_KEY
from roofai.logger import logger


NAME = module.name(__file__, NAME)


# https://app.roboflow.com/kamangir/roof-dataset-one/1/export
def download_project(
    object_name: str,
    project_name: str,
    project_version: int,
    clean: bool = True,
    verbose: bool = False,
) -> bool:
    logger.info(
        "{}.download_project: {}({}) -> {}{}".format(
            NAME,
            project_name,
            project_version,
            object_name,
            " clean" if clean else "",
        )
    )

    metadata = {
        "input": {
            "project": project_name,
            "version": project_version,
        }
    }

    object_path = objects.object_path(object_name)
    temp_path = os.path.join(
        object_path,
        "downloaded_from_roboflow",
    )
    if not path.create(temp_path, log=verbose):
        return False

    timer = ElapsedTimer()

    try:
        rf = Roboflow(api_key=ROBOFLOW_API_KEY)

        project = rf.workspace().project(project_name)

        version = project.version(project_version)

        _ = version.download(
            model_format="png-mask-semantic",
            location=temp_path,
            overwrite=True,
        )
    except Exception as e:
        logger.error(e)
        return False

    timer.stop()
    logger.info(f"took {timer.elapsed_pretty()}")

    success, classes_df = file.load_dataframe(
        os.path.join(temp_path, "train/_classes.csv"),
        log=verbose,
    )
    if not success:
        return False

    list_of_classes = [
        class_name.strip() for class_name in classes_df[" Class"].tolist()
    ]
    logger.info(
        "{} class(es): {}".format(
            len(list_of_classes),
            ", ".join(list_of_classes),
        )
    )

    for image_filename in tqdm(
        glob.glob(
            os.path.join(
                temp_path,
                "train/*.jpg",
            )
        )
    ):
        record_id = file.name(image_filename).replace(".", "-")
        logger.info(record_id)

        success, image = file.load_image(
            image_filename,
            log=verbose,
        )
        if not success:
            return False
        if not file.save_image(
            os.path.join(
                object_path,
                f"SegNet-Tutorial/CamVid/train/{record_id}.png",
            ),
            image,
            log=verbose,
        ):
            return False

        mask_filename = os.path.join(
            file.path(image_filename),
            "{}_mask.png".format(
                file.name(image_filename),
            ),
        )

        success, mask = file.load_image(
            mask_filename,
            log=verbose,
        )
        if not success:
            return False
        if not file.save_image(
            os.path.join(
                object_path,
                f"SegNet-Tutorial/CamVid/trainannot/{record_id}.png",
            ),
            mask,
            log=verbose,
        ):
            return False

        if not file.save_image(
            os.path.join(
                object_path,
                f"SegNet-Tutorial/CamVid/traina-colored/{record_id}.png",
            ),
            (
                plt.cm.viridis(mask[:, :, 0].astype(np.float32) / len(list_of_classes))
                * 255
            ).astype(np.uint8)[:, :, :3],
            log=verbose,
        ):
            return False

        if clean and not all(
            file.delete(filename_)
            for filename_ in [
                image_filename,
                mask_filename,
            ]
        ):
            return False

    if not file.save_yaml(
        os.path.join(object_path, "metadata.yaml"),
        {
            "classes": list_of_classes,
            "ingested-by": fullname(),
            "kind": "CamVid",
            "source": "gmaps",
        },
        log=True,
    ):
        return False

    return post_to_object(
        object_name,
        NAME.replace(".", "-"),
        metadata,
    )



import pytest
import numpy as np

from blue_objects import objects

from roofai import env
from roofai.tests.caching import cache_pretrainedmodels
from roofai.semseg import Profile
from roofai.google_maps.semseg.predict import predict


@pytest.mark.parametrize(
    ["lat", "lon"],
    [
        [
            env.ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LAT,
            env.ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LON,
        ],
    ],
)
def test_google_maps_predict(
    lat: float,
    lon: float,
):
    assert cache_pretrainedmodels()

    prediction_object_name = objects.unique_object("test_google_maps_predict")

    model_object_name = env.ROOFAI_DEFAULT_GOOGLE_MAPS_MODEL
    assert objects.download(model_object_name)

    success, output_image, input_image, metadata = predict(
        lat=lat,
        lon=lon,
        model_object_name=model_object_name,
        prediction_object_name=prediction_object_name,
        device="cpu",
        profile=Profile.VALIDATION,
    )
    assert success
    assert isinstance(output_image, np.ndarray)
    assert isinstance(input_image, np.ndarray)
    assert "gsd" in metadata



import matplotlib.pyplot as plt
import numpy as np
from typing import List, Any

from blueness import module
from blue_options import string
from blue_objects import file, path, objects
from blue_objects.graphics.signature import sign_filename

from roofai import NAME
from roofai.host import signature


NAME = module.name(__file__, NAME)


# helper function for data visualization
def visualize(
    images,
    filename: str = "",
    in_notebook: bool = False,
    description: List[str] = [],
    list_of_contours: List[Any] = [],
    line_width: int = 80,
    draw_the_contours_in: List[str] = ["image"],
    footer: List[str] = [],
):
    n = len(images)
    fig = plt.figure(figsize=(n * 5, 5))

    for name in images:
        if isinstance(images[name], str):
            success, images[name] = image = file.load_image(images[name], log=True)
            assert success

    for name in images:
        images[name][np.isnan(images[name])] = 0

    for i, (name, image) in enumerate(images.items()):
        ax = fig.add_subplot(1, n, i + 1)
        plt.xticks([])
        plt.yticks([])
        plt.xlabel(
            "{} - {}{}".format(
                name,
                string.pretty_shape_of_matrix(image),
                (
                    " - {} levels: {}..{}".format(
                        len(np.unique(image)),
                        int(np.min(image)),
                        int(np.max(image)),
                    )
                    if name in "prediction,mask,groundtruth".split(",")
                    else ""
                ),
            )
        )
        ax.imshow(image)

        if name in draw_the_contours_in:
            for contour in list_of_contours:
                plt.plot(
                    contour[0],
                    contour[1],
                    "o-",
                    color="orange",
                )

    if filename:
        file.prepare_for_saving(filename)
        plt.savefig(filename)
        success = sign_filename(
            filename,
            header=objects.signature(
                info=file.name(filename),
                object_name=path.name(file.path(filename)),
            )
            + description,
            footer=footer + signature(),
            line_width=line_width,
        )

    if in_notebook:
        plt.show()
    plt.close()



from typing import List
import numpy as np
from torch.utils.data import Dataset as BaseDataset

from blueness import module
from blue_objects import file, objects

from roofai.dataset.classes import DatasetTarget
from roofai.google_maps.api.static import get as get_static_image
from roofai import NAME
from roofai.logger import logger

NAME = module.name(__file__, NAME)


class GoogleMapsDataset(BaseDataset):
    def __init__(
        self,
        lat: float,
        lon: float,
        zoom: int,
        augmentation=None,
        preprocessing=None,
        count=-1,
        chip_overlap: float = 0.25,
        prediction_object_name: str = "",
        size: str = "640x640",
        verbose: bool = False,
    ):

        self.chip_height = DatasetTarget.TORCH.chip_height
        self.chip_width = DatasetTarget.TORCH.chip_width
        self.chip_overlap = chip_overlap

        success, self.matrix, metadata = get_static_image(
            lat=lat,
            lon=lon,
            filename=(
                objects.path_of(
                    object_name=prediction_object_name,
                    filename="input.png",
                )
                if verbose
                else ""
            ),
            zoom=zoom,
            maptype="satellite",
            size=size,
        )
        assert success

        self.gsd = metadata["gsd"]

        self.ids: List[str] = []
        for y in range(
            self.chip_height_offset,
            self.matrix.shape[0] - self.chip_height + self.chip_height_offset,
            self.chip_height_overlap,
        ):
            for x in range(
                self.chip_width_offset,
                self.matrix.shape[1] - self.chip_width + self.chip_width_offset,
                self.chip_width_overlap,
            ):
                self.ids += [f"{y:05d}-{x:05d}"]

        if count != -1:
            self.ids = self.ids[:count]

        self.augmentation = augmentation
        self.preprocessing = preprocessing

        logger.info(
            "{}.GoogleMapsDataset({:.05f},{:.05f}, zoom={}): {:,} chip(s) x {}x{}.".format(
                NAME,
                lat,
                lon,
                zoom,
                len(self.ids),
                self.chip_height,
                self.chip_width,
            )
        )

    @property
    def chip_width_offset(self):
        return int(
            ((self.matrix.shape[1] - self.chip_width) % self.chip_width_overlap) / 2
        )

    @property
    def chip_width_overlap(self):
        return int(self.chip_overlap * self.chip_width)

    @property
    def chip_height_offset(self):
        return int(
            ((self.matrix.shape[0] - self.chip_height) % self.chip_height_overlap) / 2
        )

    @property
    def chip_height_overlap(self):
        return int(self.chip_overlap * self.chip_height)

    def __getitem__(self, i):
        item_id = self.ids[i]  # expecting f"{y:05d}-{x:05d}"

        pieces = item_id.split("-")
        assert len(pieces) == 2
        y = int(pieces[0])
        x = int(pieces[1])

        image = self.matrix[
            y : y + self.chip_height,
            x : x + self.chip_width,
        ]

        mask = np.zeros(image.shape, dtype=np.uint8)

        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        return image, mask

    def __len__(self):
        return len(self.ids)



from blue_options.env import load_config, load_env, get_env

load_env(__name__)
load_config(__name__)


GOOGLE_MAPS_API_KEY = get_env("GOOGLE_MAPS_API_KEY")

TEST_roofAI_ingest_AIRS_v1 = get_env("TEST_roofAI_ingest_AIRS_v1")

TEST_roofAI_ingest_AIRS_v2 = get_env("TEST_roofAI_ingest_AIRS_v2")

TEST_roofAI_ingest_CamVid_v1 = get_env("TEST_roofAI_ingest_CamVid_v1")

TEST_roofAI_semseg_model_AIRS_full_v1 = get_env("TEST_roofAI_semseg_model_AIRS_full_v1")

TEST_roofAI_semseg_model_AIRS_full_v2 = get_env("TEST_roofAI_semseg_model_AIRS_full_v2")

TEST_roofAI_semseg_model_CamVid_v1 = get_env("TEST_roofAI_semseg_model_CamVid_v1")

ROOFAI_AIRS_CACHE_OBJECT_NAME = get_env("ROOFAI_AIRS_CACHE_OBJECT_NAME")

ROOFAI_TEST_GOOGLE_MAPS_LAT = get_env("ROOFAI_TEST_GOOGLE_MAPS_LAT", 0.0)
ROOFAI_TEST_GOOGLE_MAPS_LON = get_env("ROOFAI_TEST_GOOGLE_MAPS_LON", 0.0)

ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LAT = get_env("ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LAT", 0.0)
ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LON = get_env("ROOFAI_TEST_GOOGLE_MAPS_HOUSE_LON", 0.0)

ROBOFLOW_API_KEY = get_env("ROBOFLOW_API_KEY")

ROOFAI_DEFAULT_GOOGLE_MAPS_MODEL = get_env("ROOFAI_DEFAULT_GOOGLE_MAPS_MODEL")



import pytest
import numpy as np

from blue_options import string
from blue_objects import objects

from roofai import env
from roofai.google_maps.api.static import get as get_static_image


@pytest.mark.parametrize(
    ["lat"],
    [[env.ROOFAI_TEST_GOOGLE_MAPS_LAT]],
)
@pytest.mark.parametrize(
    ["lon"],
    [[env.ROOFAI_TEST_GOOGLE_MAPS_LON]],
)
def test_google_maps_get_static_image(
    lat: float,
    lon: float,
):
    object_name = objects.unique_object("test_google_maps_get_static_image")

    success, image, metadata = get_static_image(
        lat=lat,
        lon=lon,
        filename=objects.path_of(
            object_name=object_name,
            filename=f"{string.timestamp()}.png",
        ),
    )

    assert success
    assert isinstance(image, np.ndarray)
    assert len(image.shape) == 3
    assert image.shape[2] == 3

    assert "gsd" in metadata
    assert isinstance(metadata["gsd"], float)

    assert "size" in metadata



from typing import List, Tuple, Dict
import segmentation_models_pytorch as smp
import numpy as np
from tqdm import tqdm
import torch
import cv2

from blueness import module
from blue_options import string
from blue_options.elapsed_timer import ElapsedTimer
from blue_objects import objects, file
from blue_objects.mlflow.tags import get_tags
from blue_objects.metadata import post_to_object
from blue_objects.logger.matrix import log_matrix

from roofai import NAME
from roofai import fullname
from roofai.google_maps.api.geocoding import geocode
from roofai.google_maps.semseg.dataset import GoogleMapsDataset
from roofai.semseg import Profile
from roofai.semseg.augmentation import get_validation_augmentation, get_preprocessing
from roofai.semseg.model import SemSegModel
from roofai.logger import logger


NAME = module.name(__file__, NAME)


def predict(
    model_object_name: str,
    prediction_object_name: str,
    device: str,
    lat: float = 0.0,
    lon: float = 0.0,
    address: str = "",
    profile: Profile = Profile.VALIDATION,
    in_notebook: bool = False,
    batch_size: int = 32,
    verbose: bool = False,
) -> Tuple[bool, np.ndarray, np.ndarray, Dict]:
    output_matrix: np.ndarray = np.array(())
    input_matrix: np.ndarray = np.array(())

    if address:
        success, lat, lon, _ = geocode(
            address=address,
            object_name=prediction_object_name,
            verbose=verbose,
        )
        if not success:
            return success, output_matrix, input_matrix, {}

    success, model_tags = get_tags(model_object_name)
    if not success:
        return success, output_matrix, input_matrix, {}
    dataset_object_name = model_tags.get("dataset", "")
    if not dataset_object_name:
        logger.error(f"{model_object_name}.dataset not found.")
        return False, output_matrix, input_matrix, {}

    success, dataset_tags = get_tags(dataset_object_name)
    if not success:
        return success, output_matrix, input_matrix, {}
    zoom_str = dataset_tags.get("zoom", "bad-zoom-value")
    try:
        zoom = int(zoom_str)
    except Exception as e:
        logger.error(e)
        return False, output_matrix, input_matrix, {}

    model = SemSegModel(
        model_filename=objects.path_of(
            filename="model.pth",
            object_name=model_object_name,
        ),
        profile=profile,
        device=device,
    )

    logger.info(
        "{}.predict: {}{:.05f},{:.05f} zoom={} -{}-batch-size:{}-> {}".format(
            NAME,
            f'"{address}" @ ' if address else "",
            lat,
            lon,
            zoom,
            device,
            batch_size,
            prediction_object_name,
        )
    )

    preprocessing_fn = smp.encoders.get_preprocessing_fn(
        model.encoder_name,
        model.encoder_weights,
    )

    dataset = GoogleMapsDataset(
        lat=lat,
        lon=lon,
        zoom=zoom,
        augmentation=get_validation_augmentation(),
        preprocessing=get_preprocessing(preprocessing_fn),
        count=model.profile.data_count,
        prediction_object_name=prediction_object_name,
        verbose=verbose,
    )

    index_list = (
        [np.random.choice(len(dataset))]
        if model.profile == Profile.VALIDATION
        else range(len(dataset))
    )
    timer = ElapsedTimer()
    list_of_masks: List[np.ndarray] = []
    for i in tqdm(range(0, len(index_list), batch_size)):
        batch_indices = index_list[i : i + batch_size]
        images = [dataset[n][0] for n in batch_indices]

        x_tensor = torch.from_numpy(np.stack(images)).to(model.device)
        pr_masks = model.model.predict(x_tensor)
        pr_masks = pr_masks.cpu().numpy()
        list_of_masks += [pr_masks]
    timer.stop()
    logger.info(f"took {timer.elapsed_pretty()}")

    stack_of_masks = np.concatenate(list_of_masks, axis=0)

    chip_count = stack_of_masks.shape[0]
    logger.info(f"stitching {chip_count:,} chip(s)...")
    output_matrix = np.zeros(dataset.matrix.shape[:2], dtype=np.float32)
    weight_matrix = np.zeros(dataset.matrix.shape[:2], dtype=np.uint8)
    chip_index: int = 0
    for y in range(
        dataset.chip_height_offset,
        dataset.matrix.shape[0] - dataset.chip_height + dataset.chip_height_offset,
        dataset.chip_height_overlap,
    ):
        for x in range(
            dataset.chip_width_offset,
            dataset.matrix.shape[1] - dataset.chip_width + dataset.chip_width_offset,
            dataset.chip_width_overlap,
        ):
            output_matrix[
                y : y + dataset.chip_height,
                x : x + dataset.chip_width,
            ] += stack_of_masks[chip_index, 0]

            weight_matrix[
                y : y + dataset.chip_height,
                x : x + dataset.chip_width,
            ] += 1

            chip_index += 1
            if chip_index >= len(dataset):
                break
        if chip_index >= len(dataset):
            break

    output_filename = objects.path_of(
        filename="prediction.png",
        object_name=prediction_object_name,
        create=True,
    )

    weight_matrix[weight_matrix == 0] = 1  # output_matrix is zero at them anyways :)
    output_matrix = output_matrix / weight_matrix

    if not log_matrix(
        matrix=output_matrix,
        suffix=[dataset.matrix],
        header=objects.signature(
            info="prediction",
            object_name=prediction_object_name,
        )
        + ([address] if address else [])
        + [
            f"lat:{lat:.05f}",
            f"lon:{lon:.05f}",
            f"model: {model_object_name}",
            model.signature,
            f"device: {device}",
            f"profile: {profile}",
            f"batch_size: {batch_size}",
            "took {}".format(
                timer.elapsed_pretty(
                    largest=True,
                    short=True,
                )
            ),
            f"{chip_count:,} chip(s)",
            f"gsd: {dataset.gsd:.2f} m",
        ],
        footer=[fullname()],
        dynamic_range=[0, 1.0],
        filename=file.add_extension(output_filename, "png"),
        colormap=cv2.COLORMAP_JET,
        verbose=True,
    ):
        return False, output_matrix, dataset.matrix, {}

    output_matrix = output_matrix * 255
    output_matrix[output_matrix < 0] = 0
    output_matrix[output_matrix > 255] = 255
    output_matrix = output_matrix.astype(np.uint8)

    return (
        post_to_object(
            prediction_object_name,
            NAME.replace(".", "-"),
            {
                "gsd": dataset.gsd,
                "lat": lat,
                "lon": lon,
                "chip_count": chip_count,
                "creation-date": string.pretty_date(),
                "address": address,
                "elapsed_time": timer.elapsed_time,
                "model": model_object_name,
                "output_filename": file.name_and_extension(output_filename),
            },
        ),
        output_matrix,
        dataset.matrix,
        {
            "gsd": dataset.gsd,
        },
    )
